/*
 * Virtual per-process performance counters.
 *
 * Copyright (C) 1999-2000  Mikael Pettersson
 */
#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/smp_lock.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/perfctr.h>

#include "compat.h"
#include "x86.h"
#include "virtual.h"

/****************************************************************
 *								*
 * Data types and macros.					*
 *								*
 ****************************************************************/

#define DEBUG 0
#if DEBUG
#define DPRINTK(fmt,args...) printk(KERN_DEBUG __FUNCTION__ ": " fmt, ##args)
#else
#define DPRINTK(fmt,args...)
#endif

struct vperfctr {
/* User-visible fields: (must be first for mmap()) */
	struct vperfctr_state state;
/* Kernel-private fields: */
	atomic_t count;
	struct file *filp_rdwr;		/* open O_RDWR|O_CREAT|O_EXCL */
	struct file *filp_rdonly;	/* open O_RDONLY|O_CREAT */
	spinlock_t state_lock;
	unsigned state_resume_action;
	int state_new_status;		/* for RESUME_NEW */
};
#define IS_RUNNING(perfctr)	((perfctr)->state.status > 0)

#define RESUME_STOPPED	0	/* do nothing */
#define RESUME_NEW	1	/* clear counters, then enable */
#define RESUME_RUNNING	2	/* re-enable */
#define RESUME_STOP	3	/* set both actions to STOPPED */
#define RESUME_UNLINK	4	/* unlink perfctr from task */

/****************************************************************
 *								*
 * Resource management.						*
 *								*
 ****************************************************************/

/* XXX: relax this to "number of LINKED perfctrs" */
atomic_t nrctrs = ATOMIC_INIT(0);
static const char this_service[] = __FILE__;

static int inc_nrctrs(void)
{
	if( atomic_read(&nrctrs) == 0 ) {
		const char *other;
		DPRINTK("reserving hardware\n");
		if( (other = perfctr_cpu_reserve(this_service)) != NULL ) {
			printk(KERN_ERR __FILE__
			       ": cannot operate, hardware taken by '%s'\n",
			       other);
			return -EBUSY;
		}
	}
	atomic_inc(&nrctrs);
	return 0;
}

static void dec_nrctrs(void)
{
	if( atomic_dec_and_test(&nrctrs) ) {
		DPRINTK("releasing hardware\n");
		perfctr_cpu_release(this_service);
	}
}

static struct vperfctr *perfctr_alloc(void)
{
	unsigned long page;

	if( inc_nrctrs() != 0 )
		return NULL;
	page = get_zeroed_page(GFP_KERNEL);
	DPRINTK("allocated page %#lx\n", page);
	if( !page ) {
		dec_nrctrs();
		return NULL;
	}
	set_bit(PG_reserved, &mem_map[MAP_NR(page)].flags);
	MOD_INC_USE_COUNT;
	return (struct vperfctr*) page;
}

static void perfctr_free(struct vperfctr *perfctr)
{
	clear_bit(PG_reserved, &mem_map[MAP_NR(perfctr)].flags);
	free_page((unsigned long)perfctr);
	DPRINTK("freed page %#lx\n", (unsigned long)perfctr);
	dec_nrctrs();
	MOD_DEC_USE_COUNT;
}

static struct vperfctr *get_empty_perfctr(void)
{
	struct vperfctr *perfctr = perfctr_alloc();
	if( perfctr ) {
		atomic_set(&perfctr->count, 1);
		perfctr->filp_rdwr = NULL;
		perfctr->filp_rdonly = NULL;
		spin_lock_init(&perfctr->state_lock);
		perfctr->state_resume_action = RESUME_STOPPED;
	}
	return perfctr;
}

static int put_perfctr(struct vperfctr *perfctr)
{
	if( atomic_dec_and_test(&perfctr->count) ) {
		perfctr_free(perfctr);
		return 1;
	}
	return 0;
}

static __inline__ void
perfctr_accumulate(struct vperfctr *perfctr,
		   const struct perfctr_ctrs *now)
{
	int i;

	perfctr->state.sum.tsc += now->tsc - perfctr->state.start.tsc;
	for(i = perfctr->state.status; --i >= 0;)
		perfctr->state.sum.pmc[i] +=
			now->pmc[i] - perfctr->state.start.pmc[i];
}

/* Sample the CPU's counters and update state. */
static __inline__ void perfctr_sample(struct vperfctr *perfctr)
{
	struct perfctr_ctrs now;
	perfctr_cpu_read_counters(perfctr->state.status, &now);
	perfctr_accumulate(perfctr, &now);
}

/* PRE: perfctr == TASK_PERFCTR(current) && IS_RUNNING(perfctr)
 * Restart the counters.
 */
static __inline__ void perfctr_resume(struct vperfctr *perfctr)
{
	perfctr_cpu_write_control(perfctr->state.status, &perfctr->state.control);
	perfctr_cpu_read_counters(perfctr->state.status, &perfctr->state.start);
}

/****************************************************************
 *								*
 * Process management callbacks.				*
 *								*
 ****************************************************************/

/* This perfctr has just been removed from current.
 * If the counters are running, stop them and sample their final values.
 * Mark this perfctr as unlinked and decrement its use count.
 */
void __vperfctr_unlink(struct vperfctr *perfctr)
{
	DPRINTK("perfctr %p\n", perfctr);
	spin_lock(&perfctr->state_lock);
	if( IS_RUNNING(perfctr) ) {
		perfctr_sample(perfctr);
		perfctr_cpu_disable_rdpmc();
	}
	perfctr->state.status = PERFCTR_STATUS_UNLINKED;
	spin_unlock(&perfctr->state_lock);
	put_perfctr(perfctr);
}

/* schedule() --> switch_to() --> .. --> __vperfctr_suspend().
 * If the counters are running, sample their values.
 */
void __vperfctr_suspend(struct vperfctr *perfctr)
{
	if( IS_RUNNING(perfctr) ) {
		spin_lock(&perfctr->state_lock);
		perfctr_sample(perfctr);
		spin_unlock(&perfctr->state_lock);
		perfctr_cpu_disable_rdpmc();
	}
}

/* Called from __switch_to() or sys_perfctr_control().
 * PRE: perfctr == TASK_PERFCTR(current)
 * Perform the perfctr's resume action.
 */
void __vperfctr_resume(struct vperfctr *perfctr)
{
	spin_lock(&perfctr->state_lock);
	switch( perfctr->state_resume_action ) {
	case RESUME_STOPPED:
		break;
	case RESUME_NEW:
		memset(&perfctr->state.sum, 0, sizeof(perfctr->state.sum));
		perfctr->state_resume_action = RESUME_RUNNING;
		perfctr->state.status = perfctr->state_new_status;
		/*FALLTHROUGH*/
	case RESUME_RUNNING:
		perfctr_cpu_enable_rdpmc();
		perfctr_resume(perfctr);
		break;
	case RESUME_STOP:
		perfctr->state_resume_action = RESUME_STOPPED;
		perfctr->state.status = PERFCTR_STATUS_STOPPED;
		break;
	case RESUME_UNLINK:
		TASK_PERFCTR(current) = NULL;
		perfctr->state.status = PERFCTR_STATUS_UNLINKED;
		if( put_perfctr(perfctr) )
			return;
	}
	spin_unlock(&perfctr->state_lock);
}

/****************************************************************
 *								*
 * Virtual perfctr "system calls".				*
 *								*
 ****************************************************************/

static int sys_perfctr_stop(struct vperfctr *perfctr)
{
	if( perfctr != TASK_PERFCTR(current) ) {
		spin_lock(&perfctr->state_lock);
		perfctr->state_resume_action = RESUME_STOP;
		spin_unlock(&perfctr->state_lock);
	} else if( IS_RUNNING(perfctr) ) {
		spin_lock(&perfctr->state_lock);
		perfctr->state_resume_action = RESUME_STOPPED;
		perfctr->state.status = PERFCTR_STATUS_STOPPED;
		spin_unlock(&perfctr->state_lock);
		perfctr_cpu_disable_rdpmc();
	}
	return 0;
}

static int sys_perfctr_control(struct vperfctr *perfctr, void *arg)
{
	struct perfctr_control control;
	int status;

	if( copy_from_user(&control, arg, sizeof control) )
		return -EFAULT;
	if( (status = perfctr_cpu_check_control(&control)) < 0 )
		return status;
	if( status == 0 )
		return sys_perfctr_stop(perfctr);
	spin_lock(&perfctr->state_lock);
	perfctr->state.control = control;
	perfctr->state_resume_action = RESUME_NEW;
	perfctr->state_new_status = status;
	spin_unlock(&perfctr->state_lock);
	if( perfctr == TASK_PERFCTR(current) )
		__vperfctr_resume(perfctr);	/* lock-unlock */
	return 0;
}

static int sys_perfctr_unlink(struct vperfctr *perfctr)
{
	if( perfctr == TASK_PERFCTR(current) ) {
		TASK_PERFCTR(current) = NULL;
		__vperfctr_unlink(perfctr);
	} else {
		spin_lock(&perfctr->state_lock);
		perfctr->state_resume_action = RESUME_UNLINK;
		spin_unlock(&perfctr->state_lock);
	}
	return 0;
}

/****************************************************************
 *								*
 * Virtual perfctr file operations.				*
 *								*
 ****************************************************************/

static int perfctr_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct vperfctr *perfctr;
	unsigned long offset = vma_offset(vma);

	DPRINTK("filp %p, vma %p\n", filp, vma);
	DPRINTK("*vma == { start: %#lx, end: %#lx, prot: %#lx, flags: %#x, "
		"ops: %p, offset: %#lx, file: %p }\n",
		vma->vm_start, vma->vm_end, pgprot_val(vma->vm_page_prot),
		vma->vm_flags, vma->vm_ops, offset, vma->vm_file);
	/* Only allow read-only mapping of first page. */
	if( (vma->vm_end - vma->vm_start) != PAGE_SIZE ||
	    offset != 0 ||
	    (pgprot_val(vma->vm_page_prot) & _PAGE_RW) ||
	    (vma->vm_flags & (VM_WRITE | VM_MAYWRITE)) ) {
		DPRINTK("bad mapping prevented\n");
		return -EPERM;
	}
	perfctr = filp->f_dentry->d_inode->u.generic_ip;
	if( remap_page_range(vma->vm_start, virt_to_phys(perfctr),
			     PAGE_SIZE, vma->vm_page_prot) ) {
		DPRINTK("remap_page_range failed\n");
		return -EAGAIN;
	}
	DPRINTK("mapped perfctr %p to user address %p\n",
		perfctr, (void*)vma->vm_start);
	return 0;
}

/* We need to indicate if a filp is a "spectator" or a "control" port.
 * The natural solution would be to use combinations of FMODE_WRITE
 * and FMODE_READ in filp->f_mode. Unfortunately, this would allow
 * writable memory mappings to be set up on control ports.
 * Instead we mark every filp as read-only and use the
 * `private_data' field as our own read-only / read-write flag.
 */
#define MK_FILP_READONLY(filp) ((filp)->private_data = 0)
#define MK_FILP_WRITABLE(filp) ((filp)->private_data = (void*)1)
#define IS_FILP_WRITABLE(filp) ((filp)->private_data != 0)

static int perfctr_release(struct inode *inode, struct file *filp)
{
	struct vperfctr *perfctr = inode->u.generic_ip;
	DPRINTK("filp %p, perfctr %p\n", filp, perfctr);
	if( IS_FILP_WRITABLE(filp) )
		perfctr->filp_rdwr = NULL;
	else
		perfctr->filp_rdonly = NULL;
	inode->u.generic_ip = NULL;
	put_perfctr(perfctr);
	return 0;
}

static int perfctr_ioctl(struct inode *inode, struct file *filp,
			 unsigned int cmd, unsigned long arg)
{
	struct vperfctr *perfctr = inode->u.generic_ip;

	if( !IS_FILP_WRITABLE(filp) )
		return -EPERM;
	switch( cmd ) {
	case PERFCTR_VIRTUAL_CONTROL:
		return sys_perfctr_control(perfctr, (void*)arg);
	case PERFCTR_VIRTUAL_STOP:
		return sys_perfctr_stop(perfctr);
	case PERFCTR_VIRTUAL_UNLINK:
		return sys_perfctr_unlink(perfctr);
	}
	return -EINVAL;
}

static struct file_operations perfctr_file_ops = {
	.mmap = perfctr_mmap,
	.release = perfctr_release,
	.ioctl = perfctr_ioctl,
};

static struct inode_operations perfctr_inode_ops = {
	&perfctr_file_ops,
};

static struct file *get_perfctr_filp(struct vperfctr *perfctr, int writable)
{
	struct inode *inode;
	struct file *filp;

	if( (filp = get_empty_filp()) == NULL )
		return NULL;
	if( (inode = get_empty_inode()) == NULL )
		goto out_filp;
	if( (filp->f_dentry = dget(d_alloc_root(inode))) == NULL )
		goto out_inode;
	inode->i_state = I_DIRTY;
	inode->i_mode = S_IFCHR | S_IRUSR | S_IWUSR;
	inode->i_uid = current->fsuid;
	inode->i_gid = current->fsgid;
	inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	filp->f_pos = 0;
	filp->f_flags = 0;
	filp->f_mode = FMODE_READ;
	inode->u.generic_ip = perfctr;
	inode->i_op = &perfctr_inode_ops;
	filp->f_op = &perfctr_file_ops;
	if( writable ) {
		MK_FILP_WRITABLE(filp);
		perfctr->filp_rdwr = filp;
	} else {
		MK_FILP_READONLY(filp);
		perfctr->filp_rdonly = filp;
	}
	atomic_inc(&perfctr->count);
	return filp;
 out_inode:
	iput(inode);
 out_filp:
	put_filp(filp);
	return NULL;
}

static int perfctr_attach_task(struct task_struct *tsk, int writable)
{
	struct vperfctr *perfctr;
	struct file *filp;
	int ret;

	if( (perfctr = TASK_PERFCTR(tsk)) == NULL &&
	    (perfctr = get_empty_perfctr()) == NULL )
		return -ENOMEM;
	if( writable ) {
		ret = -EBUSY;
		if( perfctr->filp_rdwr )
			goto out_perfctr;
		filp = get_perfctr_filp(perfctr, 1);
	} else if( (filp = perfctr->filp_rdonly) != NULL )
		get_file(filp);
	else
		filp = get_perfctr_filp(perfctr, 0);
	ret = -ENFILE;
	if( !filp )
		goto out_perfctr;
	if( (ret = get_unused_fd()) < 0 )
		goto out_filp;
	fd_install(ret, filp);
	if( !TASK_PERFCTR(tsk) )
		TASK_PERFCTR(tsk) = perfctr;
	DPRINTK("filp %p (count %d), perfctr %p (count %d)\n",
		filp, file_count(filp),
		perfctr, atomic_read(&perfctr->count));
	return ret;
 out_filp:
	fput(filp);	/* runs filp->f_op->release on last close */
 out_perfctr:
	if( !TASK_PERFCTR(tsk) )
		perfctr_free(perfctr);
	return ret;
}

/****************************************************************
 *								*
 * /dev/perfctr ioctl commands for virtual perfctrs.		*
 *								*
 ****************************************************************/

int vperfctr_attach(int pid, int writable)
{
	struct task_struct *tsk;
	int ret;

	if( pid == current->pid )
		return perfctr_attach_task(current, writable);
	lock_kernel();
	read_lock(&tasklist_lock);
	tsk = find_task_by_pid(pid);
	read_unlock(&tasklist_lock);
	ret = -ESRCH;
	if( tsk && (tsk->flags & PF_PTRACED) &&
	    (tsk->state == TASK_STOPPED) && (tsk->p_pptr == current) )
		ret = perfctr_attach_task(tsk, writable);
	unlock_kernel();
	return ret;
}

/*
 * Sample the current process' counters and update state.
 *
 * This operation exists only to support pre-MMX Intel P5 processors,
 * since they cannot sample the counters in user-mode.
 */
int vperfctr_sample_current(void)
{
	struct vperfctr *perfctr = TASK_PERFCTR(current);
	if( perfctr && IS_RUNNING(perfctr) ) {
		struct perfctr_ctrs now;
		spin_lock(&perfctr->state_lock);
		perfctr_cpu_read_counters(perfctr->state.status, &now);
		perfctr_accumulate(perfctr, &now);
		perfctr->state.start = now;
		spin_unlock(&perfctr->state_lock);
	}
	return 0;
}

int vperfctr_unlink_current(void)
{
	struct task_struct *tsk = current;
	struct vperfctr *perfctr = TASK_PERFCTR(tsk);
	if( perfctr ) {
		TASK_PERFCTR(tsk) = NULL;
		__vperfctr_unlink(perfctr);
	}
	return 0;
}

/****************************************************************
 *								*
 * Module load/cleanup callbacks.				*
 *								*
 ****************************************************************/

#ifdef MODULE
static struct vperfctr_stub off;

void vperfctr_init(void)
{
	off = vperfctr_stub;
	vperfctr_stub.unlink = __vperfctr_unlink;
	vperfctr_stub.suspend = __vperfctr_suspend;
	vperfctr_stub.resume = __vperfctr_resume;
}

void vperfctr_cleanup(void)
{
	vperfctr_stub = off;
}
#endif
