/*
 * x86 performance-monitoring counters driver.
 *
 * Copyright (C) 1999-2000  Mikael Pettersson
 */
#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>

#include <asm/msr.h>
#include <asm/perfctr.h>

#include "compat.h"
#include "x86.h"

/****************************************************************
 *								*
 * Hardware-level declarations.					*
 *								*
 ****************************************************************/

static union {	/* Support for lazy evntsel MSR updates. */
	struct perfctr_control control;	/* 16 bytes */
	char __align[SMP_CACHE_BYTES];	/* 32 bytes */
} per_cpu_control[NR_CPUS] __cacheline_aligned;

/* Intel P5, Cyrix 6x86MX/MII, Centaur WinChip C6/2/2A/3 */
#define MSR_P5_CESR		0x11
#define MSR_P5_CTR0		0x12
#define MSR_P5_CTR1		0x13

/* Intel P6 */
#define MSR_P6_PERFCTR0		0xC1
#define MSR_P6_PERFCTR1		0xC2
#define MSR_P6_EVNTSEL0		0x186
#define MSR_P6_EVNTSEL1		0x187

/* AMD K7 Athlon */
#define MSR_K7_EVNTSEL0		0xC0010000	/* .. 0xC0010003 */
#define MSR_K7_PERFCTR0		0xC0010004	/* .. 0xC0010007 */

/****************************************************************
 *								*
 * Read TSC and PerfCtrs as `unsigned long long' values		*
 *								*
 ****************************************************************/

#define USE_A 1 /* use "=A" to assign <eax,edx> to a long long */
#define USE_DO 1 /* wrap u64 asm reads with do-while-0 loop */

#if USE_DO
#define WRAP(x,e)	do { (x) = (e); } while(0)
#else
#define WRAP(x,e)	((x) = (e))
#endif

#if USE_A

static __inline__ u64 __rdtsc64(void)
{
	u64 ull;
	rdtscll(ull);
	return ull;
}
#define rdtsc64(x)			WRAP(x,__rdtsc64())

static __inline__ u64 __rdmsr64mask(unsigned msr, unsigned himask)
{
	u64 ull;
	__asm__ __volatile__("rdmsr; andl %2,%%edx"
			     : "=A"(ull)
			     : "c"(msr), "g"(himask));
	return ull;
}
#define rdmsr64mask(x,msr,himask)	WRAP(x,__rdmsr64mask((msr),(himask)))

static __inline__ u64 __rdpmc64mask(unsigned ctr, unsigned himask)
{
	u64 ull;
	__asm__ __volatile__("rdpmc; andl %2,%%edx"
			     : "=A"(ull)
			     : "c"(ctr), "g"(himask));
	return ull;
}
#define rdpmc64mask(x,ctr,himask)	WRAP(x,__rdpmc64mask((ctr),(himask)))

#else	/* !USE_A, convert the halves via a union */

union u64_conversion {	/* safe type-punning */
	unsigned int ui[2];
	unsigned long long ull;
};

static __inline__ u64 __rdtsc64(void)
{
	union u64_conversion u;
	rdtsc(u.ui[0], u.ui[1]);
	return u.ull;
}
#define rdtsc64(x)			WRAP(x,__rdtsc64())

static __inline__ u64 __rdmsr64mask(unsigned msr, unsigned himask)
{
	union u64_conversion u;
	__asm__ __volatile__("rdmsr; andl %3,%%edx"
			     : "=a"(u.ui[0]), "=d"(u.ui[1])
			     : "c"(msr), "g"(himask));
	return u.ull;
}
#define rdmsr64mask(x,msr,himask)	WRAP(x,__rdmsr64mask((msr),(himask)))

static __inline__ u64 __rdpmc64mask(unsigned ctr, unsigned himask)
{
	union u64_conversion u;
	__asm__ __volatile__("rdpmc; andl %3,%%edx"
			     : "=a"(u.ui[0]), "=d"(u.ui[1])
			     : "c"(ctr), "g"(himask));
	return u.ull;
}
#define rdpmc64mask(x,ctr,himask)	WRAP(x,__rdpmc64mask((ctr),(himask)))

#endif	/* USE_A */

/****************************************************************
 *								*
 * Processor-specific procedures and values.			*
 *								*
 ****************************************************************/

/* Save current TSC/PMC counter values in *ctrs. */
void (*perfctr_cpu_read_counters)(int nrctrs, struct perfctr_ctrs *ctrs);

/* Write control data to the CPU's perfctr evntsel MSRs.
   This is to be done lazily: if a given evntsel MSR already
   contains the correct value, do not write to it. */
void (*perfctr_cpu_write_control)(int nrctrs, const struct perfctr_control *control);

/* Return -EPERM if the control info is invalid, otherwise the
   highest numbered enabled counter (starting from 1). */
int (*perfctr_cpu_check_control)(const struct perfctr_control *control);

/* Detected CPU type. */
unsigned char perfctr_cpu_type;

int perfctr_cpu_have_rdpmc;

static void p5_read_counters(int nrctrs, struct perfctr_ctrs *ctrs)
{
	rdtsc64(ctrs->tsc);
	rdmsr64mask(ctrs->pmc[0], MSR_P5_CTR0, 0xFF);
	if( nrctrs == 2 )
		rdmsr64mask(ctrs->pmc[1], MSR_P5_CTR1, 0xFF);
}

static void p5mmx_read_counters(int nrctrs, struct perfctr_ctrs *ctrs)
{
	rdtsc64(ctrs->tsc);
	rdpmc64mask(ctrs->pmc[0], 0, 0xFF);
	if( nrctrs == 2 )
		rdpmc64mask(ctrs->pmc[1], 1, 0xFF);
}

static void p5_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	unsigned evntsel;

	cpu = &per_cpu_control[smp_processor_id()].control;
	evntsel = control->evntsel[0];
	if( cpu->evntsel[0] != evntsel ) {
		cpu->evntsel[0] = evntsel;
		wrmsr(MSR_P5_CESR, evntsel, 0);
	}
}

static int p5_check_control(const struct perfctr_control *control)
{
	/* protect reserved and pin control bits */
	if( control->evntsel[0] & 0xFE00FE00 )
		return -EPERM;
	/* CTR1 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x00C00000 )
		return 2;
	/* CTR0 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x000000C0 )
		return 1;
	return 0;
}

static void p6_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	unsigned evntsel;

	cpu = &per_cpu_control[smp_processor_id()].control;
	if( nrctrs == 2 ) {
		evntsel = control->evntsel[1];
		if( evntsel != cpu->evntsel[1] ) {
			cpu->evntsel[1] = evntsel;
			wrmsr(MSR_P6_EVNTSEL1, evntsel, 0);
		}
	}	
	evntsel = control->evntsel[0];
	if( evntsel != cpu->evntsel[0] ) {
		cpu->evntsel[0] = evntsel;
		wrmsr(MSR_P6_EVNTSEL0, evntsel, 0);
	}
}

static int p6_check_control(const struct perfctr_control *control)
{
	/* protect reserved, interrupt control, and pin control bits */
	if( control->evntsel[0] & 0x00380000 )
		return -EPERM;
	if( control->evntsel[1] & 0x00780000 )
		return -EPERM;
	/* check global enable bit */
	if( control->evntsel[0] & 0x00400000 ) {
		/* check CPL field */
		if( control->evntsel[1] & 0x00030000 )
			return 2;
		if( control->evntsel[0] & 0x00030000 )
			return 1;
	}
	return 0;
}

static void k7_read_counters(int nrctrs, struct perfctr_ctrs *ctrs)
{
	int i;

	rdtsc64(ctrs->tsc);
	for(i = nrctrs; --i >= 0;)
		rdpmc64mask(ctrs->pmc[i], i, 0xFFFF);
}

static void k7_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	int i;
	unsigned evntsel;

	cpu = &per_cpu_control[smp_processor_id()].control;
	for(i = nrctrs; --i >= 0;) {
		evntsel = control->evntsel[i];
		if( evntsel != cpu->evntsel[i] ) {
			cpu->evntsel[i] = evntsel;
			wrmsr(MSR_K7_EVNTSEL0+i, evntsel, 0);
		}
	}
}

static int k7_check_control(const struct perfctr_control *control)
{
	int i, last_on;

	last_on = -1;
	for(i = 0; i < 4; ++i) {
		/* protect reserved, interrupt control, and pin control bits */
		if( control->evntsel[i] & 0x00380000 )
			return -EPERM;
		/* check enable bit */
		if( !(control->evntsel[i] & 0x00400000) )
			continue;
		/* check CPL field */
		if( control->evntsel[i] & 0x00030000 )
			last_on = i;
	}
	return last_on + 1;
}

static void mii_read_counters(int nrctrs, struct perfctr_ctrs *ctrs)
{
	rdtsc64(ctrs->tsc);
	rdpmc64mask(ctrs->pmc[0], 0, 0xFFFF);
	if( nrctrs == 2 )
		rdpmc64mask(ctrs->pmc[1], 1, 0xFFFF);
}

static int mii_check_control(const struct perfctr_control *control)
{
	/* protect reserved and pin control bits */
	if( control->evntsel[0] & 0xFA00FA00 )
		return -EPERM;
	/* CTR1 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x00C00000 )
		return 2;
	/* CTR0 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x000000C0 )
		return 1;
	return 0;
}

#if defined(CONFIG_PERFCTR_WINCHIP)
static void c6_read_counters(int nrctrs, struct perfctr_ctrs *ctrs)
{
	++ctrs->tsc;
	rdpmc64mask(ctrs->pmc[0], 0, 0xFF);
	rdpmc64mask(ctrs->pmc[1], 1, 0xFF);
}

static int c6_check_control(const struct perfctr_control *control)
{
	/* protect reserved bits */
	if( control->evntsel[0] & 0xFF00FF00 )
		return -EPERM;
	/* we have to assume both counters are active */
	return 2;
}
#endif	/* CONFIG_PERFCTR_WINCHIP */

/****************************************************************
 *								*
 * Initialisation procedures.					*
 *								*
 ****************************************************************/

#ifdef CONFIG_PERFCTR_INIT_TESTS

#define NITER	64
#define X2(S)	S";"S
#define X8(S)	X2(X2(X2(S)))

static void __init do_rdpmc(unsigned unused1, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("rdpmc") : : "c"(0) : "eax", "edx");
}

static void __init do_rdmsr(unsigned msr, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("rdmsr") : : "c"(msr) : "eax", "edx");
}

static void __init do_wrmsr(unsigned msr, unsigned data)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("wrmsr") : : "c"(msr), "a"(data), "d"(0));
}

static void __init do_rdcr4(unsigned unused1, unsigned unused2)
{
	unsigned i;
	unsigned dummy;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("movl %%cr4,%0") : "=r"(dummy));
}

static void __init do_wrcr4(unsigned cr4, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__(X8("movl %0,%%cr4") : : "r"(cr4));
}

static void __init do_empty_loop(unsigned unused1, unsigned unused2)
{
	unsigned i;
	for(i = 0; i < NITER/8; ++i)
		__asm__ __volatile__("" : : "c"(0));
}

static unsigned __init run(void (*doit)(unsigned, unsigned),
			   unsigned arg1, unsigned arg2)
{
	unsigned start, dummy, stop;
	rdtsc(start, dummy);
	(*doit)(arg1, arg2);	/* should take < 2^32 cycles to complete */
	rdtsc(stop, dummy);
	return stop - start;
}

static struct {
	const char *name;
	unsigned ticks;
} access_methods[] __initdata = {
	{ "rdpmc" },
	{ "rdmsr (counter)" },
	{ "rdmsr (evntsel)" },
	{ "wrmsr (counter)" },
	{ "wrmsr (evntsel)" },
	{ "read %cr4" },
	{ "write %cr4" },
	{ "loop overhead" },
};

static void __init init_tests_message(void)
{
	printk("Please email the following PERFCTR INIT lines "
	       "to mikpe@csd.uu.se\n"
	       "To remove this message, reconfigure "
	       "with CONFIG_PERFCTR_INIT_TESTS=n\n");
	printk("PERFCTR INIT: vendor %u, family %u, model %u\n",
	       boot_cpu_data.x86_vendor,
	       boot_cpu_data.x86,
	       boot_cpu_data.x86_model);
}

static void __init
measure_overheads(unsigned msr_evntsel0, unsigned evntsel0, unsigned msr_perfctr0)
{
	int i;

	access_methods[0].ticks = perfctr_cpu_have_rdpmc ? run(do_rdpmc,0,0) : 0;
	access_methods[1].ticks = run(do_rdmsr, msr_perfctr0, 0);
	access_methods[2].ticks = run(do_rdmsr, msr_evntsel0, 0);
	access_methods[3].ticks = run(do_wrmsr, msr_perfctr0, 0);
	access_methods[4].ticks = run(do_wrmsr, msr_evntsel0, evntsel0);
	access_methods[5].ticks = run(do_rdcr4, 0, 0);
	access_methods[6].ticks = run(do_wrcr4, get_cr4(), 0);
	access_methods[7].ticks = run(do_empty_loop, 0, 0);
	wrmsr(msr_evntsel0, 0, 0);

	init_tests_message();
	printk("PERFCTR INIT: NITER == %u\n", NITER);
	for(i = 0; i < ARRAY_SIZE(access_methods); ++i) {
		printk("PERFCTR INIT: %s ticks == %u\n",
		       access_methods[i].name, access_methods[i].ticks);
	}
}

static void __init p5_init_tests(void)
{
	unsigned evnt = 0x16 | (3 << 6);
	measure_overheads(MSR_P5_CESR, evnt, MSR_P5_CTR0);
}

static void __init p6_init_tests(void)
{
	unsigned evnt = 0xC0 | (3 << 16) | (1 << 22);
	measure_overheads(MSR_P6_EVNTSEL0, evnt, MSR_P6_PERFCTR0);
}

static void __init k7_stop_and_clear(void)
{
	wrmsr(MSR_K7_EVNTSEL0+0, 0, 0);
	wrmsr(MSR_K7_EVNTSEL0+1, 0, 0);
	wrmsr(MSR_K7_PERFCTR0+0, 0, 0);
	wrmsr(MSR_K7_PERFCTR0+1, 0, 0);
}

static unsigned __init k7_test(unsigned ctr0_on, unsigned ctr1_on)
{
	unsigned ctr1_evnt = 0xC0 | (3 << 16);	/* num insns */
	unsigned ctr0_evnt = 0xDF | (3 << 16);	/* dr3 match */
	const unsigned ENable = (1<<22);
	unsigned dummy, ctr1_result;

	if( ctr1_on ) ctr1_evnt |= ENable;
	if( ctr0_on ) ctr0_evnt |= ENable;
	wrmsr(MSR_K7_EVNTSEL0+1, ctr1_evnt, 0);
	wrmsr(MSR_K7_EVNTSEL0+0, ctr0_evnt, 0);
	do_empty_loop(0, 0);
	rdmsr(MSR_K7_PERFCTR0+1, ctr1_result, dummy);
	k7_stop_and_clear();
	return ctr1_result;
}

static void __init k7_check_how_ENable_works(void)
{
	unsigned test0, test1, test2, test3;

	k7_stop_and_clear();

	test0 = k7_test(0, 0); /* should be == 0 */
	test1 = k7_test(1, 1); /* should be > 0 */
	test2 = k7_test(0, 1); /* if > 0, EvntSel0 does not override */
	test3 = k7_test(1, 0); /* if > 0, only EvntSel0 has ENable */

	printk("PERFCTR INIT: Athlon test0 == %u (%s)\n",
	       test0, test0 == 0 ? "ok" : "unexpected");
	printk("PERFCTR INIT: Athlon test1 == %u (%s)\n",
	       test1, test1 > 0 ? "ok" : "unexpected");
	printk("PERFCTR INIT: Athlon test2 == %u (%s)\n",
	       test2, test2 > 0 ? "EvntSel0 does not override" : "ok");
	printk("PERFCTR INIT: Athlon test3 == %u (%s)\n",
	       test3, test3 > 0 ? "only EvntSel0 has ENable" : "ok");
}

static void __init k7_init_tests(void)
{
	unsigned evnt = 0xC0 | (3 << 16) | (1 << 22);
	measure_overheads(MSR_K7_EVNTSEL0, evnt, MSR_K7_PERFCTR0);
	k7_check_how_ENable_works();
}

#define mii_init_tests()	p5_init_tests()

#if defined(CONFIG_PERFCTR_WINCHIP)
static void __init c6_init_tests(void)
{
	unsigned int cesr, dummy;

	rdmsr(MSR_P5_CESR, cesr, dummy);
	init_tests_message();
	printk("PERFCTR INIT: boot CESR == %#08x\n", cesr);
}
#endif

#else	/* !CONFIG_PERFCTR_INIT_TESTS */
#define p5_init_tests()
#define p6_init_tests()
#define k7_init_tests()
#define mii_init_tests()
#define c6_init_tests()
#endif	/* CONFIG_PERFCTR_INIT_TESTS */

static int __init intel_init(void)
{
	if( !(boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 5:
		if( boot_cpu_data.x86_capability & X86_FEATURE_MMX ) {
			perfctr_cpu_type = PERFCTR_CPU_INTEL_P5MMX;
			perfctr_cpu_have_rdpmc = 1;
			perfctr_cpu_read_counters = p5mmx_read_counters;
		} else {
			perfctr_cpu_type = PERFCTR_CPU_INTEL_P5;
			perfctr_cpu_have_rdpmc = 0;
			perfctr_cpu_read_counters = p5_read_counters;
		}
		perfctr_cpu_write_control = p5_write_control;
		perfctr_cpu_check_control = p5_check_control;
		p5_init_tests();
		return 0;
	case 6:
		if( boot_cpu_data.x86_model >= 7 )	/* PIII */
			perfctr_cpu_type = PERFCTR_CPU_INTEL_PIII;
		else if( boot_cpu_data.x86_model >= 3 )	/* PII or Celeron */
			perfctr_cpu_type = PERFCTR_CPU_INTEL_PII;
		else
			perfctr_cpu_type = PERFCTR_CPU_INTEL_P6;
		perfctr_cpu_have_rdpmc = 1;
		perfctr_cpu_read_counters = p5mmx_read_counters;
		perfctr_cpu_write_control = p6_write_control;
		perfctr_cpu_check_control = p6_check_control;
		p6_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init amd_init(void)
{
	if( !(boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* K7 Athlon
		 * Model 2 adds local APIC.
		 */
		perfctr_cpu_type = PERFCTR_CPU_AMD_K7;
		perfctr_cpu_have_rdpmc = 1;
		perfctr_cpu_read_counters = k7_read_counters;
		perfctr_cpu_write_control = k7_write_control;
		perfctr_cpu_check_control = k7_check_control;
		k7_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init cyrix_init(void)
{
	if( !(boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* 6x86MX or MII */
		perfctr_cpu_type = PERFCTR_CPU_CYRIX_MII;
		perfctr_cpu_have_rdpmc = 1;
		perfctr_cpu_read_counters = mii_read_counters;
		perfctr_cpu_write_control = p5_write_control;
		perfctr_cpu_check_control = mii_check_control;
		mii_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init centaur_init(void)
{
#if defined(CONFIG_PERFCTR_WINCHIP)
	switch( boot_cpu_data.x86 ) {
	case 5:
		switch( boot_cpu_data.x86_model ) {
		case 4: /* WinChip C6 */
			perfctr_cpu_type = PERFCTR_CPU_WINCHIP_C6;
			break;
		case 8: /* WinChip 2 or 2A */
		case 9: /* WinChip 3, a 2A with 128KB L1 cache */
			perfctr_cpu_type = PERFCTR_CPU_WINCHIP_2;
			break;
		default:
			return -ENODEV;
		}
		/*
		 * TSC must be inaccessible for perfctrs to work.
		 */
		if( !(get_cr4() & X86_CR4_TSD) ||
		    (boot_cpu_data.x86_capability & X86_FEATURE_TSC) )
			return -ENODEV;
		perfctr_cpu_have_rdpmc = 1;
		perfctr_cpu_read_counters = c6_read_counters;
		perfctr_cpu_write_control = p5_write_control;
		perfctr_cpu_check_control = c6_check_control;
		c6_init_tests();
		return 0;
	}
#endif
	return -ENODEV;
}

int __init perfctr_cpu_init(void)
{
	int err = -ENODEV;
	if( boot_cpu_data.x86_capability & X86_FEATURE_MSR ) {
		switch( boot_cpu_data.x86_vendor ) {
		case X86_VENDOR_INTEL:
			err = intel_init();
			break;
		case X86_VENDOR_AMD:
			err = amd_init();
			break;
		case X86_VENDOR_CYRIX:
			err = cyrix_init();
			break;
		case X86_VENDOR_CENTAUR:
			err = centaur_init();
		}
	}
	if( err )
		return err;
	/*
	 * We initialise per_cpu_control[] to contain "impossible"
	 * evntsel values guaranteed to differ from anything accepted
	 * by perfctr_cpu_check_control(). This way, initialisation of
	 * a CPUs evntsel MSRs will happen automatically the first time
	 * perfctr_cpu_write_control() executes on it.
	 * All-bits-one works for all currently supported processors.
	 * [XXX: Should be set up by the cpu-specific init procedure.]
	 */
	memset(per_cpu_control, ~0, sizeof per_cpu_control);

	return 0;
}

/****************************************************************
 *								*
 * Hardware reservation.					*
 *								*
 ****************************************************************/

static const char *current_service = 0;

const char *perfctr_cpu_reserve(const char *service)
{
	if( current_service )
		return current_service;
	current_service = service;
	return 0;
}

void perfctr_cpu_release(const char *service)
{
	if( service != current_service ) {
		printk(KERN_ERR __FUNCTION__
		       ": attempt by %s to release while reserved by %s\n",
		       service, current_service);
	} else
		current_service = 0;
}
